//////////////////////////////////////////////////////////////////////////////
// Mirror-DTC JNI Library for Ubuntu
// 
// (C) 2011-2012 T.Ishii (t-ishii@js2.so-net.ne.jp) All rights Reserved.
//////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <gdk/gdkx.h>
#include <pulse/simple.h>
#include <pulse/error.h>
#include <pulse/gccmacro.h>
#include <X11/keysym.h>
#include <X11/extensions/XTest.h>
#include <jni.h>

typedef unsigned int DWORD;
typedef int LONG;
typedef short WORD;
typedef unsigned char BYTE;

#define MAXMONITOR 16

#define INFINITE   0x7fffffff

#define F_SCREEN   0x0001
#define F_CURSOR   0x0002
#define F_SOUND    0x0004
#define F_AERO     0x0008
#define F_KEYBOARD 0x0010
#define F_POSITION 0x0020
#define F_MOUSE    0x0040

////////////////////////////////////
// Global Variable
////////////////////////////////////

// for Screen
int g_monitornum;
GdkRectangle g_monitorrect[MAXMONITOR+2];
GdkWindow *g_rootwindow;
int g_screenimgbufwidth;
int g_screenimgbufheight;
// for Cursor
int g_cursorwidth = 0;
int g_cursorheight = 0;
int g_cursorhotspotx = 0;
int g_cursorhotspoty = 0;
DWORD *g_CursorImgBuf = NULL;
// for Audio
pa_simple *g_pulseaudio = NULL;
int g_audiobuflen = 0;
BYTE *g_AudioBuf = NULL;
// for Position
Display *g_xdisplay;
int g_xscreen;
Window g_xwindow;
// for Keyboard
int g_KeyBind[2*256];

////////////////////////////////////
// Sub Function
////////////////////////////////////
void SetCursorInfo()
{
  GdkCursor *cursor;
  GdkPixbuf *cursorimg;
  unsigned char *pixel;
  int rowstride;
  int x,y;
  unsigned char *bmpbuf;

  cursor = gdk_cursor_new_for_display(gdk_display_get_default(),GDK_LEFT_PTR);
  cursorimg = gdk_cursor_get_image(cursor);
  sscanf(gdk_pixbuf_get_option(cursorimg,"x_hot"),"%d",&g_cursorhotspotx);
  sscanf(gdk_pixbuf_get_option(cursorimg,"y_hot"),"%d",&g_cursorhotspoty);
  g_cursorwidth = gdk_pixbuf_get_width(cursorimg);
  g_cursorheight = gdk_pixbuf_get_height(cursorimg);
  pixel = gdk_pixbuf_get_pixels(cursorimg);
  rowstride = gdk_pixbuf_get_rowstride(cursorimg);
  g_CursorImgBuf = malloc(g_cursorwidth*g_cursorheight*sizeof(DWORD));
  bmpbuf = (unsigned char *)g_CursorImgBuf;
  for (y = 0; y < g_cursorheight; y++) {
    for (x = 0; x < g_cursorwidth; x++) {
      bmpbuf[x*4+2+y*g_cursorwidth*4] = pixel[x*4+0+y*rowstride];
      bmpbuf[x*4+1+y*g_cursorwidth*4] = pixel[x*4+1+y*rowstride];
      bmpbuf[x*4+0+y*g_cursorwidth*4] = pixel[x*4+2+y*rowstride];
      bmpbuf[x*4+3+y*g_cursorwidth*4] = pixel[x*4+3+y*rowstride];
    }
  }
  g_object_unref(cursorimg);
}

////////////////////////////////////
// Interface Function
////////////////////////////////////
JNIEXPORT jint JNICALL Java_MCSrvioJNI_getSupportVer(JNIEnv *env, jobject obj,jbyteArray asciiname)
{
  BYTE *name = (BYTE *)"JNI Library for Ubuntu (Ver 1.1)";

  // Set Ascii JNI Library Name (max 64byte)
  (*env)->SetByteArrayRegion(env,asciiname,0,36+1,name);
  // Return Interface Version
  return 1;
}

JNIEXPORT jint JNICALL Java_MCSrvioJNI_getSupportFunc(JNIEnv *env, jobject obj)
{
  // Initialize 
  //  g_thread_init(NULL);
  gdk_threads_init();
  //  XInitThreads();
  gdk_init(NULL,NULL);
  g_xdisplay = gdk_x11_display_get_xdisplay(gdk_display_get_default());
  g_xscreen = XDefaultScreen(g_xdisplay);
  g_xwindow = RootWindow(g_xdisplay,g_xscreen);
  g_rootwindow = gdk_screen_get_root_window(gdk_screen_get_default());
  SetCursorInfo();
  // Return Support Function Mask
  return F_SCREEN | F_CURSOR | F_SOUND | F_KEYBOARD | F_POSITION | F_MOUSE;
}

//////////////////////////////////////////////
// Screen Function (F_SCREEN)
//////////////////////////////////////////////
JNIEXPORT int JNICALL Java_MCSrvioJNI_makeScreenInfo(JNIEnv *env, jobject obj)
{
  GdkScreen *screen;
  int primary;
  int i;
  int left,top,right,bottom;
  GdkDisplay *disp;

  // Get Info from Current Default Display
  disp = gdk_display_open(NULL);
  screen = gdk_display_get_screen(disp,0);
  g_monitornum = gdk_screen_get_n_monitors(screen);
  if (g_monitornum > MAXMONITOR) {
    g_monitornum = MAXMONITOR;
  }
  primary = gdk_screen_get_primary_monitor(screen);
  if (primary > MAXMONITOR) {
    primary = 0;
  }
  for (i = 0; i < g_monitornum; i++) {
    gdk_screen_get_monitor_geometry(screen,i,&g_monitorrect[1+i]);
  }
  gdk_display_close(disp);
  // Set Primary Monitor Rect
  g_monitorrect[0].x = g_monitorrect[1+primary].x;
  g_monitorrect[0].y = g_monitorrect[1+primary].y;
  g_monitorrect[0].width = g_monitorrect[1+primary].width;
  g_monitorrect[0].height = g_monitorrect[1+primary].height;
  // Set Composite Monitor Rect
  if (g_monitornum > 1) {
    left = INFINITE;
    top = INFINITE;
    right = -INFINITE;
    bottom = -INFINITE;
    for (i = 0; i < g_monitornum; i++) {
      if (g_monitorrect[1+i].x < left) {
	left = g_monitorrect[1+i].x;
      }
      if (g_monitorrect[1+i].y < top) {
	top = g_monitorrect[1+i].y;
      }
      if ((g_monitorrect[1+i].x+g_monitorrect[1+i].width) > right) {
	right = (g_monitorrect[1+i].x+g_monitorrect[1+i].width);
      }
      if ((g_monitorrect[1+i].y+g_monitorrect[1+i].height) > bottom) {
	bottom = (g_monitorrect[1+i].y+g_monitorrect[1+i].height);
      }
    }
    g_monitorrect[1+g_monitornum].x = left;
    g_monitorrect[1+g_monitornum].y = top;
    g_monitorrect[1+g_monitornum].width = right-left;
    g_monitorrect[1+g_monitornum].height = bottom-top;
  }
  return g_monitornum;
}

JNIEXPORT void JNICALL Java_MCSrvioJNI_getScreenRect(JNIEnv *env, jobject obj,jint no,jintArray info)
{
  jint carr[4];

  carr[0] = g_monitorrect[no].x;
  carr[1] = g_monitorrect[no].y;
  carr[2] = g_monitorrect[no].width;
  carr[3] = g_monitorrect[no].height;
  (*env)->SetIntArrayRegion(env,info,0,4,(const jint *)carr);
}

JNIEXPORT void JNICALL Java_MCSrvioJNI_makeImageBuf(JNIEnv *env, jobject obj,jint width,jint height)
{
  g_screenimgbufwidth = width;
  g_screenimgbufheight = height;
}

JNIEXPORT void JNICALL Java_MCSrvioJNI_getScreenImage(JNIEnv *env, jobject obj,jintArray argbdata,jint no,jint posx,jint posy,jint width,jint height) {
  jint *carr;
  int rowstride;
  int x,y;
  XImage *image;
  unsigned long xpixel;
  unsigned long *lp;
  int screenwidth,screenheight;
  GdkDisplay *disp;
  Display *xdisplay;

  disp = gdk_display_open(NULL);
  xdisplay = gdk_x11_display_get_xdisplay(disp);
  if ((posx+width) > g_monitorrect[no].width) {
    width = g_monitorrect[no].width-posx;
  }
  if ((posy+height) > g_monitorrect[no].height) {
    height = g_monitorrect[no].height-posy;
  }
  image = XGetImage(xdisplay,GDK_ROOT_WINDOW(),g_monitorrect[no].x+posx,g_monitorrect[no].y+posy,width,height,0xffffffff,ZPixmap);
  gdk_display_close(disp);
  if (image == NULL) {
    return;
  }
  for (y = 0; y < height; y++) {
    (*env)->SetIntArrayRegion(env,argbdata,posx+(posy+y)*g_screenimgbufwidth,width,&(((int *)image->data)[y*width]));
  }    
  XDestroyImage(image);
}

////////////////////////////////////////
// Cursor Data Function (F_CURSOR)
////////////////////////////////////////
JNIEXPORT jint JNICALL Java_MCSrvioJNI_getCursorData(JNIEnv *env, jobject obj,jintArray info,jintArray argbdata,jbyteArray xorflg)
{
  jint retval = 0;// if set xorflg valkd. (xorflg is only useful in Windows)
  jint carr[4];

  // Set Info
  carr[0] = g_cursorwidth;
  carr[1] = g_cursorheight;
  carr[2] = g_cursorhotspotx;
  carr[3] = g_cursorhotspoty;
  (*env)->SetIntArrayRegion(env,info,0,4,carr);
  // Set Image
  if (g_cursorwidth > 128 || g_cursorheight > 128) {
    int *work = malloc(128*128*sizeof(int));
    int width = 128;
    int height = 128;
    int x,y;

    for (y = 0; y < 128; y++) {
      for (x = 0; x < 128; x++) {
	work[x+y*128] = g_CursorImgBuf[x+y*g_cursorwidth];
      }
    }
    (*env)->SetIntArrayRegion(env,argbdata,0,width*height,work);
    free(work);
  } else {
    (*env)->SetIntArrayRegion(env,argbdata,0,g_cursorwidth*g_cursorheight,g_CursorImgBuf);
  }
  return retval;
}

//////////////////////////////////////////
// Sound Function (F_SOUND)
//////////////////////////////////////////
JNIEXPORT int JNICALL Java_MCSrvioJNI_startRec(JNIEnv *env, jobject obj,jint freq,jint ch,jint bits)
{
  pa_sample_spec ss;
  int error;

  ss.rate = freq;
  ss.channels = ch;
  if (bits != 8) {
    ss.format = PA_SAMPLE_S16LE;
  } else {
    ss.format = PA_SAMPLE_U8;
  }
  if (!(g_pulseaudio = pa_simple_new(NULL, "MirrorDTC Server", PA_STREAM_RECORD, NULL, "record", &ss, NULL, NULL, &error))) {
    return -1;
  }
  return 0;
}

JNIEXPORT jint JNICALL Java_MCSrvioJNI_getRecData(JNIEnv *env, jobject obj,jbyteArray data,jint len)
{
  int error;

  if (g_audiobuflen < len) {
    if (g_AudioBuf) {
      free(g_AudioBuf);
    }
    g_AudioBuf = malloc(len);
    g_audiobuflen = len;
  }
  if (pa_simple_read(g_pulseaudio,g_AudioBuf,len,&error) < 0) {
    return -1;
  }
  (*env)->SetByteArrayRegion(env,data,0,len,g_AudioBuf);
  return len;
}

JNIEXPORT void JNICALL Java_MCSrvioJNI_endRec(JNIEnv *env, jobject obj)
{
  if (g_pulseaudio == NULL) {
    return;
  }
  pa_simple_free(g_pulseaudio);
  g_pulseaudio = NULL;
}

/////////////////////////////////////////////
// Keyboard Function (F_KEYBOARD)
/////////////////////////////////////////////
JNIEXPORT void JNICALL Java_MCSrvioJNI_setKeyBind(JNIEnv *env, jobject obj,jintArray keybind)
{
  (*env)->GetIntArrayRegion(env,keybind,0,512,g_KeyBind);
}

JNIEXPORT void JNICALL Java_MCSrvioJNI_keyPress(JNIEnv *env, jobject obj,jint scancode)
{
  int i;
  int keycode;

  for (i = 0; i < 256; i++) {
    if (g_KeyBind[2*i] == scancode) {
      break;
    }
    if (g_KeyBind[2*i] == 0x7fffffff) {
      return;
    }
  }
  if (i == 256) {
    return;
  }
  keycode = g_KeyBind[2*i+1];
  XTestFakeKeyEvent(g_xdisplay,XKeysymToKeycode(g_xdisplay,keycode),True,0);
}

JNIEXPORT void JNICALL Java_MCSrvioJNI_keyRelease(JNIEnv *env, jobject obj,jint scancode)
{
  int i;
  int keycode;

  for (i = 0; i < 256; i++) {
    if (g_KeyBind[2*i] == scancode) {
      break;
    }
    if (g_KeyBind[2*i] == 0x7fffffff) {
      return;
    }
  }
  if (i == 256) {
    return;
  }
  keycode = g_KeyBind[2*i+1];
  XTestFakeKeyEvent(g_xdisplay,XKeysymToKeycode(g_xdisplay,keycode),False,0);
}

//////////////////////////////////////////////
// Current Position Function (F_POSITION)
//////////////////////////////////////////////
JNIEXPORT void JNICALL Java_MCSrvioJNI_getCursorPos(JNIEnv *env, jobject obj,jintArray pos)
{
  Window root_return,child_return;
  int root_x_return,root_y_return,win_x_return,win_y_return,mask_return;
  int jpos[2];

  XQueryPointer(g_xdisplay,g_xwindow,&root_return,&child_return,&root_x_return,&root_y_return,&win_x_return,&win_y_return,&mask_return);
  jpos[0] = root_x_return;
  jpos[1] = root_y_return;
  (*env)->SetIntArrayRegion(env,pos,0,2,jpos);
}

/////////////////////////////////////////////
// Mouse Function (F_MOUSE)
/////////////////////////////////////////////
JNIEXPORT void JNICALL Java_MCSrvioJNI_mouseMove(JNIEnv *env, jobject obj,jint x,jint y)
{
  XTestFakeMotionEvent(g_xdisplay,g_xscreen,x,y,0);
}

JNIEXPORT void JNICALL Java_MCSrvioJNI_mouseButton(JNIEnv *env, jobject obj,jint no,jint mode)
{
  if (no < 0 || no >= 3) {
    return;
  }
  if (mode != 0) {
    XTestFakeButtonEvent(g_xdisplay,1+no,True,0);
  } else {
    XTestFakeButtonEvent(g_xdisplay,1+no,False,0);
  }
}

JNIEXPORT void JNICALL Java_MCSrvioJNI_mouseWheel(JNIEnv *env, jobject obj,jint delta)
{
  if (delta > 0) {
    XTestFakeButtonEvent(g_xdisplay,1+3,True,0);
    XTestFakeButtonEvent(g_xdisplay,1+3,False,0);
  } else {
    XTestFakeButtonEvent(g_xdisplay,1+4,True,0);
    XTestFakeButtonEvent(g_xdisplay,1+4,False,0);
  }
}
